/*
 * meye_control.c - Routines for controlling two mechanical eyes.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h> // for the sleep(2)
#include "aserial.h"
#include "meye_control.h"

#define EYER_VERT_ID	"00"
#define EYER_VERT_PIN	6
#define EYER_VERT_MID	77 // bigger is rotate up
#define EYER_MAX_UP		(EYER_VERT_MID+45)
#define EYER_MAX_DOWN	(EYER_VERT_MID-27)
#define EYER_HORIZ_ID	"01"
#define EYER_HORIZ_PIN	7
#ifdef __linux__
#define EYER_HORIZ_MID	97 // bigger rotates Izzy's left
#endif
#ifdef __QNXNTO__
#define EYER_HORIZ_MID  97 //121 // bigger rotates Izzy's left
#endif
#define EYER_MAX_RIGHT	(EYER_HORIZ_MID-45)
#define EYER_MAX_LEFT	(EYER_HORIZ_MID+45)
#define EYEL_VERT_ID	"02"
#define EYEL_VERT_PIN	10
#define EYEL_VERT_MID	112 // bigger is rotate down
#define EYEL_MAX_UP		(EYEL_VERT_MID-27)
#define EYEL_MAX_DOWN	(EYEL_VERT_MID+45)
#define EYEL_HORIZ_ID	"03"
#define EYEL_HORIZ_PIN	11
#define EYEL_HORIZ_MID	113 // bigger rotates Izzy's left
#define EYEL_MAX_RIGHT	(EYEL_HORIZ_MID-45)
#define EYEL_MAX_LEFT	(EYEL_HORIZ_MID+45)

static double eye_center_d = 33.85; // distance in millimeters from the center 
									// of each eye to the center a the spot 
									// between them i.e. half the distance 
									/// between the eyes

meyec_t *meyec_new(void)
{
	meyec_t *ec = NULL;
	meye_t *me;
#ifdef __linux__
	char *serialport = "/dev/ttyACM0";
#endif
#ifdef __QNXNTO__
	char *serialport = "/dev/serusb1";
#endif
	char buf[100];
	ssize_t n;

	if ((ec = malloc(sizeof(meyec_t))) == NULL)
		goto errorout;

	if ((ec->as = aserial_new()) == NULL)
		goto errorout;

	if (aserial_open(ec->as, serialport, 115200) == -1)
		goto errorout;

	// NEEDED???@@@ tcdrain((*ec->as)->fd);
	
	me = &ec->e[MEYE_R];
	strcpy(me->vert_id, EYER_VERT_ID);
	me->vert_cur = 0;
	me->vert_pin = EYER_VERT_PIN;
	me->vert_mid = EYER_VERT_MID;
	me->max_up = EYER_MAX_UP;
	me->max_down = EYER_MAX_DOWN;
	me->sign_ud = -1;
	strcpy(me->horiz_id, EYER_HORIZ_ID);
	me->horiz_cur = 0;
	me->horiz_pin = EYER_HORIZ_PIN;
	me->horiz_mid = EYER_HORIZ_MID;
	me->max_right = EYER_MAX_RIGHT;
	me->max_left = EYER_MAX_LEFT;

	me = &ec->e[MEYE_L];
	strcpy(me->vert_id, EYEL_VERT_ID);
	me->vert_cur = 0;
	me->vert_pin = EYEL_VERT_PIN;
	me->vert_mid = EYEL_VERT_MID;
	me->max_up = EYEL_MAX_UP;
	me->max_down = EYEL_MAX_DOWN;
	me->sign_ud = -1;
	strcpy(me->horiz_id, EYEL_HORIZ_ID);
	me->horiz_cur = 0;
	me->horiz_pin = EYEL_HORIZ_PIN;
	me->horiz_mid = EYEL_HORIZ_MID;
	me->max_right = EYEL_MAX_RIGHT;
	me->max_left = EYEL_MAX_LEFT;

	me = &ec->e[MEYE_R];

	snprintf(buf, 100, "cr=%s,%d", me->vert_id, me->vert_pin);
sleep(2);
	if ((n = aserial_write(ec->as, buf, strlen(buf))) == -1) {
		perror("aserial_write()");
		goto errorout;
	}
	if (n != strlen(buf)) {
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
		fprintf(stderr, "requested=%lu, written=%ld\n", strlen(buf), n);
#else // 32-bit
		fprintf(stderr, "requested=%u, written=%d\n", strlen(buf), n);
#endif
		goto errorout;
	}

	snprintf(buf, 100, "cr=%s,%d", me->horiz_id, me->horiz_pin);
	if ((n = aserial_write(ec->as, buf, strlen(buf))) == -1) {
		perror("aserial_write()");
		goto errorout;
	}
	if (n != strlen(buf)) {
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
		fprintf(stderr, "requested=%lu, written=%ld\n", strlen(buf), n);
#else // 32-bit
		fprintf(stderr, "requested=%u, written=%d\n", strlen(buf), n);
#endif
		goto errorout;
	}

	me = &ec->e[MEYE_L];

	snprintf(buf, 100, "cr=%s,%d", me->vert_id, me->vert_pin);
	if ((n = aserial_write(ec->as, buf, strlen(buf))) == -1) {
		perror("aserial_write()");
		goto errorout;
	}
	if (n != strlen(buf)) {
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
		fprintf(stderr, "requested=%lu, written=%ld\n", strlen(buf), n);
#else // 32-bit
		fprintf(stderr, "requested=%u, written=%d\n", strlen(buf), n);
#endif
		goto errorout;
	}

	snprintf(buf, 100, "cr=%s,%d", me->horiz_id, me->horiz_pin);
	if ((n = aserial_write(ec->as, buf, strlen(buf))) == -1) {
		perror("aserial_write()");
		goto errorout;
	}
	if (n != strlen(buf)) {
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
		fprintf(stderr, "requested=%lu, written=%ld\n", strlen(buf), n);
#else // 32-bit
		fprintf(stderr, "requested=%u, written=%d\n", strlen(buf), n);
#endif
		goto errorout;
	}

	return ec;

errorout:
	if (ec) {
		if (ec->as) {
			aserial_close(ec->as);
			aserial_destroy(ec->as);
		}
		free(ec);
	}
	return NULL;
}

void meyec_destroy(meyec_t *ec)
{
	if (ec) {
		if (ec->as) {
			aserial_close(ec->as);
			aserial_destroy(ec->as);
		}
		free(ec);
	}
}

double meyec_get_eye_spacing_half(void)
{
	return eye_center_d;
}

int meyec_move_eye_angle(meyec_t *ec, int meye, int orient, int angle)
{
	size_t nbytes;
	char buf[100], *id;
	ssize_t n;
	int do_nothing = 0;

	printf("meye=%s,orient=%s,angle=%d\n",
		meye==MEYE_R?"MEYE_R":"MEYE_L", orient==MEYE_V?"MEYE_V":"MEYE_H",
		angle);
	if (orient == MEYE_V) {
		id = ec->e[meye].vert_id;
		if (meye == MEYE_R) {
			if (angle > ec->e[meye].max_up)
				angle = ec->e[meye].max_up;
			else if (angle < ec->e[meye].max_down)
				angle = ec->e[meye].max_down;
		} else if (meye == MEYE_L) {
			if (angle < ec->e[meye].max_up)
				angle = ec->e[meye].max_up;
			else if (angle > ec->e[meye].max_down)
				angle = ec->e[meye].max_down;
		}
		if (angle == ec->e[meye].vert_cur)
			do_nothing = 1;
	} else if (orient == MEYE_H) {
		id = ec->e[meye].horiz_id;
		if (meye == MEYE_R) {
			if (angle > ec->e[meye].max_left)
				angle = ec->e[meye].max_left;
			else if (angle < ec->e[meye].max_right)
				angle = ec->e[meye].max_right;
		} else if (meye == MEYE_L) {
			if (angle > ec->e[meye].max_left)
				angle = ec->e[meye].max_left;
			else if (angle < ec->e[meye].max_right)
				angle = ec->e[meye].max_right;
		}
		if (angle == ec->e[meye].horiz_cur)
			do_nothing = 1;
	} else {
		fprintf(stderr, "meyec_move_angle(): invalid orient %d\n", orient);
		return -1;
	}
	if (!do_nothing) {
		snprintf(buf, 100, "ma=%s,%d", id, angle);
		nbytes = strlen(buf);
		if ((n = aserial_write(ec->as, buf, nbytes)) == -1) {
			perror("aserial_write()");
			return -1;
		}
		if (n != nbytes) {
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
			fprintf(stderr, "requested=%lu, written=%ld\n", nbytes, n);
#else // 32-bit
			fprintf(stderr, "requested=%u, written=%d\n", nbytes, n);
#endif
			return -1;
		}
		if (orient == MEYE_V) 
			ec->e[meye].vert_cur = angle;
		else
			ec->e[meye].horiz_cur = angle;
		aserial_read(ec->as, buf, sizeof(buf));
	}
	if (do_nothing) {
		printf("meyec_move_eye_angle() doing nothing: angle=%d\n", angle);
	}
	return 0;
}

int meyec_look_straight_ahead(meyec_t *ec)
{
	meye_t *me;

	me = &ec->e[MEYE_R];
	if (meyec_move_eye_angle(ec, MEYE_R, MEYE_V, me->vert_mid) == -1)
		return -1;
	if (meyec_move_eye_angle(ec, MEYE_R, MEYE_H, me->horiz_mid) == -1)
		return -1;
	me = &ec->e[MEYE_L];
	if (meyec_move_eye_angle(ec, MEYE_L, MEYE_V, me->vert_mid) == -1)
		return -1;
	if (meyec_move_eye_angle(ec, MEYE_L, MEYE_H, me->horiz_mid) == -1)
		return -1;
	return 0;
}

#ifndef __QNXNTO__
/*
 * meyec_calc_eye_angles - Given a point in space somewhere relative to a point 
 * halfway between the two eyes, this calculates the servo motor angles
 * needed for the two eyes to both look at the point in space.
 * Parameters:
 * Ha - Horizontal angle from a point halfway between the two eyes to
 * the desired point in space.
 * Va - Vertical angle from a point halfway between the two eyes to
 * the desired point in space.
 * rld - Distance in millimeters from a point halfway between the two 
 * eyes to the desired point in space.
 * RHa, RVa - Horizontal and vertical angles in degrees for the servo 
 * motors for the right eye.
 * LHa, LVa - Horizontal and vertical angles in degrees for the servo 
 * motors for the right eye.
 */
void meyec_calc_eye_angles(meyec_t *ec, double *RHa, double *RVa, double *LHa, double *LVa, double Ha, double Va, double rld)
{
	double ad, dd, HBa = 180-Ha; //, VBa = 180-Va;

	meye_t *mer, *mel;

	mer = &ec->e[MEYE_R];
	mel = &ec->e[MEYE_L];

	// Making use of the law of Cosines to first find the unknown length 
	// of a side so the we can then use a rearranged version of the law
	// of cosines to calculate a desired angle.
  
	//printf("Ha = %f\n", Ha);
	ad = sqrt(rld*rld + eye_center_d*eye_center_d - (2 * rld * eye_center_d * cos(Ha*(M_PI/180))));
	//printf("ad = %f\n", ad);
	*RHa = acos((eye_center_d*eye_center_d + ad*ad - rld*rld)/(2 * eye_center_d * ad))*(180/M_PI);
	//printf("Raw RHa = %f\n", &RHa);
	dd = sqrt(rld*rld + eye_center_d*eye_center_d - (2 * rld * eye_center_d * cos(HBa*(M_PI/180))));
	//printf("Raw dd = %f\n", dd);
	*LHa = acos((eye_center_d*eye_center_d + dd*dd - rld*rld)/(2 * eye_center_d * dd))*(180/M_PI);
	//printf("Raw LHa = %f\n", *LHa);

	// Eyes straight ahead is .horiz_mid which is somewhere near halfway between 
	// 0 and 180 where 0 is all the way to the right and 180 is all the way to the left.
	// calc_eye_angles() is giving angles where 0 is also all the way to the right and
	// for the right eye is calculating an angle where 0 is all the way to the left and
	// for the left eye is calculating an angle where 0 is all the way to the right.
	// But .horiz_mid isn't exactly 90 and could be on either side of it.
	// e.g. if eyer.horiz_mid=99 for the right eye and Ra=93.54 then we need to use the
	// angle on the other side of the sightline, which is 180-Ra=86.46. The .horiz_mid 
	// descrepancy is 90-99=-9 and so we should use Ra-(-9)=95.46.
	// e.g. if eyel.horiz_mid=80 for the left eye and La=73.78 then the .horiz_mid 
	// descrepancy is 90-80=10 and so we should use La-10=63.78.
  
	printf("Before compensating for horiz_mid, *RHa=%f, mer->horiz_mid=%d\n",
			 *RHa, mer->horiz_mid);
	*RHa = (180-*RHa)-(90-mer->horiz_mid);
	//printf("Adjusted Ra = %f\n", *Ra);
	printf("Before compensating for horiz_mid, *LHa=%f, mel->horiz_mid=%d\n",
			 *LHa, mel->horiz_mid);
	*LHa = *LHa-(90-mel->horiz_mid);
	//printf("Adjusted La = %f\n", *La);

	// Vertical is tricky because for the right eye,
	// all the way down is 0 degrees and all the way up is 180 degrees
	// but for the left eye,
	// all the way up is 0 degrees and all the way down is 180 degrees.
	// The vertical value we're given is all the way down is 0 and all the way up is 180.
  
	*RVa = Va-(90-mer->vert_mid);
	*LVa = (180-Va)-(90-mel->vert_mid);
}

int meyec_move_eyes_to_loc(meyec_t *ec, double Ha, double Va, double rld)
{
	double RHa, RVa, LHa, LVa;

	meyec_calc_eye_angles(ec, &RHa, &RVa, &LHa, &LVa, Ha, Va, 300);
	if (meyec_move_eye_angle(ec, MEYE_R, MEYE_H, RHa) == -1)
		return -1;
	if (meyec_move_eye_angle(ec, MEYE_R, MEYE_V, RVa) == -1)
		return -1;
	if (meyec_move_eye_angle(ec, MEYE_L, MEYE_H, LHa) == -1)
		return -1;
	if (meyec_move_eye_angle(ec, MEYE_L, MEYE_V, LVa) == -1)
		return -1;
	return 0;
}
#endif

